from collections.abc import Iterable
import csv

from django.contrib.auth.mixins import PermissionRequiredMixin
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.http import HttpResponse
from django.views.generic import ListView

from wafer.talks.models import Talk

from bursary.models import Bursary
from front_desk.models import CheckIn
from invoices.models import Invoice
from register.models import Accomm, Attendee, ChildCare, Meal, Visa


class CSVExportView(ListView):
    """Export the given columns for the model as CSV."""
    columns = None
    filename = None

    def get_data_line(self, instance):
        ret = []
        for column in self.columns:
            obj = instance
            for component in column.split('.'):
                try:
                    obj = getattr(obj, component)
                except ObjectDoesNotExist:
                    obj = '%s missing!' % component
                    break
                except AttributeError:
                    obj = getattr(self, component)(obj)
                if not obj:
                    break
                if callable(obj):
                    obj = obj()
            if (not isinstance(obj, (str, bytes))
                    and isinstance(obj, Iterable)):
                ret.extend(str(i) for i in obj)
            else:
                ret.append(str(obj))

        return ret

    def write_rows(self, writer, objects):
        for instance in objects:
            writer.writerow(self.get_data_line(instance))

    def render_to_response(self, context, **response_kwargs):
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = (
            'attachment; filename="%s"' % self.filename
        )

        writer = csv.writer(response)
        writer.writerow(self.columns)
        self.write_rows(writer, context['object_list'])

        return response


class AttendeeAdminMixin(PermissionRequiredMixin):
    permission_required = 'register.change_attendee'


class AttendeeBadgeExport(AttendeeAdminMixin, CSVExportView):
    name = 'Attendee badges'
    model = Attendee
    filename = "attendee_badges.csv"
    ordering = ('user__username',)
    columns = [
        'user.username', 'reconfirm', 'user.email', 'user.get_full_name',
        'nametag_2', 'nametag_3', 'languages', 'food.diet',
    ]


class AttendeeAccommExport(AttendeeAdminMixin, CSVExportView):
    name = 'Accommodation'
    model = Accomm
    filename = "attendee_accommodation.csv"
    ordering = ('attendee__user__username',)
    columns = [
        'attendee.user.username', 'attendee.user.get_full_name',
        'attendee.user.email', 'attendee.reconfirm', 'attendee.paid',
        'attendee.user.bursary.accommodation_status', 'attendee.gender',
        'attendee.country', 'option', 'requirements',
        'family_usernames', 'get_checkin_checkouts', 'room',
    ]


class ChildCareExport(AttendeeAdminMixin, CSVExportView):
    name = 'Childcare'
    model = ChildCare
    filename = "attendee_child_care.csv"
    ordering = ('attendee__user__username',)
    columns = [
        'attendee.user.username', 'attendee.user.get_full_name',
        'attendee.user.email', 'attendee.reconfirm', 'attendee.paid',
        'attendee.user.bursary.accommodation_status',
        'attendee.arrival', 'attendee.departure',
        'needs', 'details',
    ]


class TalksExport(PermissionRequiredMixin, CSVExportView):
    name = 'Talks'
    model = Talk
    permission_required = 'talks.edit_private_notes'
    filename = "talk_evaluations.csv"
    ordering = ("talk_id",)
    columns = [
        'talk_id', 'title', 'get_authors_display_name', 'abstract',
        'talk_type.name', 'track.name', 'get_status_display', 'review_score',
        'review_count', 'notes', 'private_notes', 'all_review_comments',
    ]

    def all_review_comments(self, talk):
        return [
            "(%s) %s" % (review.reviewer.username, review.notes.raw)
            for review in talk.reviews.all()
            if review.notes.raw
        ]


class FoodExport(AttendeeAdminMixin, CSVExportView):
    name = 'Food'
    model = Meal
    filename = "meals.csv"
    ordering = ('date', 'meal',)
    columns = [
        'date', 'meal', 'total', 'total_unconfirmed',
        'regular', 'vegetarian', 'vegan', 'gluten_free',
        'other', 'other_details',
    ]

    def render_to_response(self, context, **response_kwargs):
        self._confirmed_attendees = {}
        return super().render_to_response(context, **response_kwargs)

    def attendee_confirmed(self, attendee_id):
        if attendee_id not in self._confirmed_attendees:
            if CheckIn.objects.filter(attendee_id=attendee_id).exists():
                return True

            attendee = Attendee.objects.get(id=attendee_id)
            paid = attendee.paid()

            try:
                bursary = Bursary.objects.get(user=attendee.user)
            except Bursary.DoesNotExist:
                bursary = Bursary()

            confirmed = any((
                not bursary.request_any and attendee.billable() and paid,
                not bursary.request_any and not attendee.billable() and attendee.final_dates,
                bursary.request_any and bursary.status_in(None, ['accepted']),
                attendee.reconfirm,
            ))
            self._confirmed_attendees[attendee_id] = confirmed
        return self._confirmed_attendees[attendee_id]

    def get_data_line(self, meal):
        row = {
            'date': meal.date.isoformat(),
            'meal': meal.meal,
            'total': 0,
            'total_unconfirmed': meal.food_set.count(),
            'regular': 0,
            'vegetarian': 0,
            'vegan': 0,
            'gluten_free': 0,
            'other': 0,
            'other_details': [],
        }

        for food in meal.food_set.all():
            if not self.attendee_confirmed(food.attendee_id):
                continue

            diet = food.diet
            if diet == '':
                diet = 'regular'
            elif diet == 'other':
                details = [food.attendee.user.username]
                details.append(food.special_diet)
                row['other_details'].append(': '.join(details))
            row[diet] += 1
            row['total'] += 1

        row['other_details'] = ', '.join(row['other_details'])
        return [row.get(key) for key in self.columns]


class SpecialDietExport(AttendeeAdminMixin, CSVExportView):
    name = 'Special diets'
    filename = "diets.csv"
    ordering = ('attendee__user__username',)
    columns = [
        'username', 'name', 'confirmed', 'diet', 'special_diet'
    ]

    def attendee_confirmed(self, attendee):
        try:
            if attendee.check_in:
                return True
        except CheckIn.DoesNotExist:
            pass

        if attendee.reconfirm:
            return True

        paid = attendee.paid()

        try:
            bursary = attendee.user.bursary
        except Bursary.DoesNotExist:
            bursary = Bursary()

        return any((
            not bursary.request_any and attendee.billable() and paid,
            not bursary.request_any and not attendee.billable() and attendee.final_dates,
            bursary.request_any and bursary.status_in(None, ['accepted']),
        ))

    def get_queryset(self):
        return Meal.objects.get(**self.kwargs).food_set.exclude(
                special_diet__exact='',
                diet__exact='',
            ).select_related(
                'attendee',
                'attendee__user',
                'attendee__user__bursary',
                'attendee__user__userprofile',
            ).order_by(*self.ordering)

    def get_data_line(self, food):
        row = {
            'username': food.attendee.user.username,
            'name': food.attendee.user.userprofile.display_name(),
            'confirmed': self.attendee_confirmed(food.attendee),
            'diet': food.diet,
            'special_diet': food.special_diet,
        }
        return [row.get(key) for key in self.columns]

    def write_rows(self, writer, objects):
        super().write_rows(writer, objects)
        count = {True: 0, False: 0}
        for food in Meal.objects.get(**self.kwargs).food_set.filter(
                special_diet__exact='',
                diet__exact='',
            ).select_related(
                'attendee',
                'attendee__user',
                'attendee__user__userprofile',
            ):
            count[self.attendee_confirmed(food.attendee)] += 1
        for confirmed in (True, False):
            row = {
                'username': '*',
                'name': 'Everyone Else - {} people'.format(count[confirmed]),
                'confirmed': confirmed,
                'diet': '',
                'special_diet': '',
            }
            writer.writerow([row.get(key) for key in self.columns])


class BursaryExport(AttendeeAdminMixin, CSVExportView):
    name = 'Bursaries'
    filename = "bursaries.csv"

    columns = [
        'user.username',
        'user.userprofile.display_name',
        'user.email',
        'user.attendee.arrived',
        'user.attendee.country',
        'travel_from',
        'travel_status',
        'food_status',
        'accommodation_status',
        'expenses_status',
    ]

    approved = ('pending', 'accepted',)

    def get_queryset(self):
        return Bursary.objects.filter(
            Q(travel_status__in=self.approved)
            | Q(food_status__in=self.approved)
            | Q(accommodation_status__in=self.approved)
            | Q(expenses_status__in=self.approved)
        ).exclude(
            user__attendee__id=None, # exclude unregistered people
        ).prefetch_related(
            'user',
            'user__attendee',
            'user__userprofile',
        ).order_by('user__username')


class FingerprintExport(AttendeeAdminMixin, CSVExportView):
    name = 'Fingerprints'
    filename = 'fingerprints.csv'

    queryset = Attendee.objects.exclude(pgp_fingerprints='')
    columns = (
        'user.username',
        'user.userprofile.display_name',
        'user.email',
        'pgp_fingerprints',
        'keysigning_id',
    )


class InvoiceExport(AttendeeAdminMixin, CSVExportView):
    name = 'Invoices'
    filename = 'invoices.csv'

    queryset = Invoice.objects.all()
    ordering = ('reference_number',)
    columns = (
        'reference_number',
        'status',
        'date',
        'last_update',
        'recipient.username',
        'recipient.userprofile.display_name',
        'compound',
        'transaction_id',
        'total',
    )


class VisaExport(AttendeeAdminMixin, CSVExportView):
    name = 'Visas'
    filename = 'visas.csv'

    queryset = Visa.objects.all()
    columns = (
        'attendee.user.username',
        'attendee.user.userprofile.display_name',
        'attendee.user.email',
        'country',
        'request_travel_bursary',
        'travel_bursary_status',
    )

    def request_travel_bursary(self, visa):
        try:
            return visa.attendee.user.bursary.request_travel
        except ObjectDoesNotExist:
            return False

    def travel_bursary_status(self, visa):
        try:
            return visa.attendee.user.bursary.travel_status
        except ObjectDoesNotExist:
            return ''



# We have no models, but some model-related behaviour

import debconf.talk_urls
from wafer.schedule.admin import SCHEDULE_ITEM_VALIDATORS

# Workaround strictness of wafer schedule validation
non_contiguous = None
for i, validator in enumerate(SCHEDULE_ITEM_VALIDATORS):
    _, err_type, _ = validator
    if err_type == "non_contiguous":
        non_contiguous = i
        break
if non_contiguous is not None:
    SCHEDULE_ITEM_VALIDATORS.pop(non_contiguous)



from django.conf.urls import url

from volunteers.views import (
    TaskView, TasksView, VideoMassScheduleView, VideoShirtView,
    VolunteerStatisticsView, VolunteerView, VolunteerUpdate,
)

urlpatterns = [
    url(r'^$', TasksView.as_view(), name='wafer_tasks'),
    url(r'^tasks/(?P<pk>\d+)/$', TaskView.as_view(), name='wafer_task'),
    url(r'^statistics/$', VolunteerStatisticsView.as_view(),
        name='wafer_volunteer_statistics'),
    url(r'^(?P<slug>[\w.@+-]+)/$', VolunteerView.as_view(),
        name='wafer_volunteer'),
    url(r'^(?P<slug>[\w.@+-]+)/update/$', VolunteerUpdate.as_view(),
        name='wafer_volunteer_update'),
    url(r'^admin/video_mass_schedule/$', VideoMassScheduleView.as_view(),
        name='wafer_volunteer_video_mass_schedule'),
    url(r'^admin/video_shirts/$', VideoShirtView.as_view(),
        name='wafer_volunteer_video_shirt'),
]



from sys import exit
from django.core.management.base import BaseCommand

from wafer.schedule.admin import SCHEDULE_ITEM_VALIDATORS
from wafer.schedule.admin import SLOT_VALIDATORS
from wafer.schedule.admin import prefetch_schedule_items
from wafer.schedule.admin import prefetch_slots


class Command(BaseCommand):
    help = 'Validates the schedule and displays any issues'

    def handle(self, *args, **options):
        # This is a copy of wafer.schedule.admin.validate_schedule as submitted
        # in https://github.com/CTPUG/wafer/pull/468
        #
        # After/when that is merged and available in wafer, we can replace the
        # code below with a simple loop over the return value of
        # wafer.schedule.admin.validate_schedule.

        rc = 0

        all_items = prefetch_schedule_items()
        for validator, _type, msg in SCHEDULE_ITEM_VALIDATORS:
            for item in validator(all_items):
                print('%s: %s' % (msg, item))
                rc = 1

        all_slots = prefetch_slots()
        for validator, _type, msg in SLOT_VALIDATORS:
            for slot in validator(all_slots):
                print('%s: %s' % (msg, slot))
                rc = 1

        exit(rc)



from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import JsonResponse
from django.utils.crypto import constant_time_compare
from django.utils import timezone
from django.views.generic import View
from register.models import Attendee
from front_desk.models import CheckIn


class DCScheduleArrived(View):
    def arrived_users(self):
        queryset = Attendee.objects.filter(announce_me=True).select_related(
            'user',
            'user__userprofile',
            'check_in',
        )
        now = timezone.now()
        for attendee in queryset:
            user = attendee.user
            try:
                arrived = attendee.check_in is not None
            except CheckIn.DoesNotExist:
                arrived = False

            departed = False
            if attendee.departure:
                departed = attendee.departure < now

            yield {
                'username': user.username,
                'arrived': arrived,
                'departed': departed,
                'name': user.userprofile.display_name(),
                'nick': attendee.nametag_3,
            }

    def get(self, request, *args, **kwargs):
        auth_header = request.META.get('HTTP_AUTHORIZATION', '')
        auth = auth_header.split(None, 1)
        if not constant_time_compare(auth, ['Bearer', settings.DCSCHEDULE_TOKEN]):
            raise PermissionDenied('Missing/Invalid Authorization token')
        return JsonResponse({
            'arrived': list(self.arrived_users())
        })





from django.conf import settings
from crispy_forms.layout import Field
from django_countries import Countries

from debconf.countries import EXTRA_COUNTRIES
from register.dates import conference_dinner_meal, parse_date


class MealSelectionField(Field):
    template = 'register/fields/meals.html'

    def render(self, form, form_style, context, **kwargs):
        field = self.fields[0]
        bound_field = form[field]

        by_day = []
        last_day = None
        day_meals = []
        for widget in bound_field.subwidgets:
            data = widget.data.copy()
            if data['value'] == conference_dinner_meal():
                data['conference_dinner'] = True

            meal, date = data['value'].split('_')
            date = parse_date(date)
            if date != last_day:
                day_meals = []
                by_day.append((date, day_meals))
                last_day = date

            day_meals.append(data)

        context['DEBCONF_BREAKFAST'] = settings.DEBCONF_BREAKFAST
        context['by_day'] = by_day
        return super().render(form, form_style, context, **kwargs)


class NightSelectionField(Field):
    template = 'register/fields/nights.html'


class RequiredCountries(Countries):
    first = ('',)
    override = {**EXTRA_COUNTRIES, **{
        '': 'Not Selected',
    }}


class OptionalCountries(Countries):
    first = ('__',)
    override = {**EXTRA_COUNTRIES, **{
        '__': 'Decline to state',
    }}



from django import forms

from crispy_forms.helper import FormHelper


class BillingForm(forms.Form):
    invoiced_entity = forms.CharField(
        label='Company',
        help_text='If you want your company to be named on the invoice, '
                  'name them here. '
                  'If blank, the invoice will be made out to you.',
        required=False,
    )
    billing_address = forms.CharField(
        label='My billing address',
        widget=forms.Textarea(attrs={'rows': 3}),
        required=False,
    )

    def __init__(self, *args, **kwargs):
        attendee = kwargs.pop('attendee')
        super().__init__(*args, **kwargs)

        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False

        self.billable = attendee.billable()

    def clean(self):
        cleaned_data = super().clean()
        if self.billable and not cleaned_data.get('billing_address'):
            self.add_error('billing_address',
                           'Paid attendees need to provide a billing address')



from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.forms import Form

from bursary.models import Bursary
from register.models import Attendee, Food
from register.models.accommodation import accomm_option_choices
from register.views.core import RegisterStep


class ReviewView(RegisterStep):
    title = 'Review'
    form_class = Form
    template_name = 'register/page/review.html'

    def get_context_data(self, **kwargs):
        user = self.request.user

        try:
            attendee = user.attendee
        except Attendee.DoesNotExist:
            attendee = None

        try:
            bursary = user.bursary
        except Bursary.DoesNotExist:
            bursary = None

        try:
            food = user.attendee.food
        except ObjectDoesNotExist:
            food = None

        try:
            accomm = user.attendee.accomm
        except ObjectDoesNotExist:
            accomm = None

        try:
            child_care = user.attendee.child_care
        except ObjectDoesNotExist:
            child_care = None

        try:
            visa = user.attendee.visa
        except ObjectDoesNotExist:
            visa = None

        context = super().get_context_data(**kwargs)

        context.update({
            'DEBCONF_SHOE_SIZES': settings.DEBCONF_SHOE_SIZES,
            'DEBCONF_BURSARY_CURRENCY': settings.DEBCONF_BURSARY_CURRENCY,
            'DEBCONF_BURSARY_CURRENCY_SYMBOL':
                settings.DEBCONF_BURSARY_CURRENCY_SYMBOL,
            'RECONFIRMATION': settings.RECONFIRMATION,
            'accomm': accomm,
            'attendee': attendee,
            'bursary': bursary,
            'child_care': child_care,
            'food': food,
            'profile': user.userprofile,
            'user': user,
            'visa': visa,
        })

        if accomm and accomm.option:
            context.update({
                'accomm_option': dict(accomm_option_choices())[accomm.option],
            })
        if attendee:
            context.update({
                'fee': Attendee.FEES[attendee.fee],
                'gender': Attendee.GENDERS[attendee.gender],
                't_shirt_size':
                    dict(settings.DEBCONF_T_SHIRT_SIZES)[attendee.t_shirt_size],
            })
        if bursary:
            context['bursary_need'] = Bursary.BURSARY_NEEDS.get(bursary.need)
        if food:
            context['diet'] = Food.DIETS[food.diet]

        return context



from exports.views import VisaExport
from register.models import Visa


class FakeWriter:
    """Like a CSV DictWriter, for tests"""
    def __init__(self):
        self.rows = []

    def writerow(self, data):
        self.rows.append(data)


def get_exported_data(cbv):
    instance = cbv()
    object_list = instance.get_queryset()
    context = instance.get_context_data(object_list=object_list)
    writer = FakeWriter()
    instance.write_rows(writer, context['object_list'])
    return [dict(zip(instance.columns, row)) for row in writer.rows]


def test_visa_export_empty(attendee):
    data = get_exported_data(VisaExport)
    assert data == []


def test_visa_export_no_bursary(attendee):
    Visa.objects.create(attendee=attendee, country='TT')
    data = get_exported_data(VisaExport)
    assert data == [{
        'attendee.user.username': attendee.user.username,
        'attendee.user.userprofile.display_name':
            attendee.user.userprofile.display_name(),
        'attendee.user.email': attendee.user.email,
        'country': 'TT',
        'request_travel_bursary': 'False',
        'travel_bursary_status': '',
    }]


def test_visa_export_submitted_bursary(bursary):
    attendee = bursary.user.attendee
    Visa.objects.create(attendee=attendee, country='TT')
    data = get_exported_data(VisaExport)
    assert data == [{
        'attendee.user.username': attendee.user.username,
        'attendee.user.userprofile.display_name':
            attendee.user.userprofile.display_name(),
        'attendee.user.email': attendee.user.email,
        'country': 'TT',
        'request_travel_bursary': 'True',
        'travel_bursary_status': 'submitted',
    }]



from django.conf import settings
from django.contrib.auth.models import Group
from django.contrib.sites.models import Site
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import EmailMultiAlternatives
from django.core.management.base import BaseCommand

from wafer.talks.models import Talk


FROM = 'content@debconf.org'
SUBJECT = '%(conference)s - talk scheduled: %(title)s'
BODY = '''\
Dear speaker / BoF convenor,

We are glad to announce that we have just published the %(conference)s schedule:

https://%(site)s/schedule/

Your Talk/BoF titled %(title)s
has been scheduled at %(time)s in %(venue)s.

We kindly request that you check whether you are able to be there at the day
and time we have allocated to your session. If not, please let us know, by
replying to this email, as soon as possible. We can either reschedule you in a
more convenient time, or if you cannot join us, we will be able to mark the
slot as available for other activities.

Best regards,
The DebConf Content Team
'''


class Command(BaseCommand):
    help = "Notify speakers that their talks have been scheduled"

    def add_arguments(self, parser):
        parser.add_argument('--subject', type=str, metavar='SUBJECT',
                            help='String to use as subject template for the emails'),
        parser.add_argument('--body', type=open, metavar='BODY',
                            help='File with template for the email body'),
        parser.add_argument('--yes', action='store_true',
                            help='Actually do something'),
        parser.add_argument('--again', action='store_true',
                            help='Mail all speakers, even those who were already contacted'),

    def badger(self, talk, dry_run, again, subject_template, body_template):
        try:
            scheduleitem = talk.scheduleitem_set.get()
        except ObjectDoesNotExist:
            return

        kv, created = talk.kv.get_or_create(
            group=self.content_group,
            key='notified_speaker',
            defaults={'value': None},
        )

        rebadger_key = [scheduleitem.venue.id, scheduleitem.slots.first().id]
        if kv.value == rebadger_key and not again:
            return

        to = [user.email for user in talk.authors.all()]

        site = Site.objects.first().domain
        subst = {
            'title': talk.title,
            'time': scheduleitem.get_start_time(),
            'venue': scheduleitem.venue.name,
            'conference': settings.DEBCONF_NAME,
            'site': site,
        }

        subject = subject_template % subst
        body = body_template % subst

        if dry_run:
            print('I would badger speakers of: %s'
                  % talk.title)
            return
        email_message = EmailMultiAlternatives(
            subject, body, from_email=FROM, to=to)
        email_message.send()

        kv.value = rebadger_key
        kv.save()

    def handle(self, *args, **options):
        dry_run = not options['yes']
        again = options['again']
        self.content_group = Group.objects.get_by_natural_key('Talk Mentors')

        if dry_run:
            print('Not actually doing anything without --yes')

        subject = options.get("subject") or SUBJECT
        body = options["body"] and options["body"].read() or BODY
        for talk in Talk.objects.all():
            self.badger(talk, dry_run, again, subject, body)



from django import forms
from django.conf import settings

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Field, Fieldset, Layout, HTML

from register.dates import nights, night_choices
from register.fields import NightSelectionField
from register.models import Accomm
from register.models.accommodation import (
        PAID_ACCOMMODATION, accomm_option_choices)
from bursary.models import Bursary


class AccommodationForm(forms.Form):
    accomm = forms.BooleanField(
        label='I need conference-organised accommodation',
        widget=forms.Select(choices=(
            (False, 'No, I will find my own accommodation'),
            (True, 'Yes, I need accommodation'),
        )),
        required=False,
    )
    nights = forms.MultipleChoiceField(
        label="I'm requesting accommodation for these nights:",
        help_text='The "night of" is the date of the day before a night. '
                  'So accommodation on the night of 6 Aug ends on the '
                  'morning of the 7th.',
        choices=night_choices(),
        widget=forms.CheckboxSelectMultiple,
        required=False,
    )
    option = forms.ChoiceField(
        label="Which accommodation would you like?",
        help_text='Different accommodation options may have different costs. '
                  'See: <a href="/about/accommodation">here</a> for details.',
        choices=accomm_option_choices(),
        required=False,
    )
    requirements = forms.CharField(
        label='Do you have any particular accommodation requirements?',
        help_text='Anything that you want us to consider for room attribution '
                  'should be listed here (ex. "I want to be with Joe Hill", '
                  '"I snore", "I go to bed early", "I need wheelchair access")',
        widget=forms.Textarea(attrs={'rows': 3}),
        required=False,
    )
    family_usernames = forms.CharField(
        label='Usernames of my family members, '
              'who have registered separately',
        help_text="One per line. This isn't validated.",
        widget=forms.Textarea(attrs={'rows': 3}),
        required=False,
    )

    def __init__(self, *args, **kwargs):
        attendee = kwargs.pop('attendee')
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False

        try:
            self.bursary = attendee.user.bursary
        except Bursary.DoesNotExist:
            self.bursary = Bursary()

        try:
            self.orga_nights = bool(
                (set(nights(orga=True)) - set(nights()))
                & set(night.date for night in attendee.accomm.nights.all())
            )
        except Accomm.DoesNotExist:
            self.orga_nights = False

        if PAID_ACCOMMODATION:
            prices = [option['price'] for option
                      in settings.PRICES['accomm'].values()
                      if 'price' in option]
            cheapest = min(prices)
            from_ = 'from ' if len(set(prices)) > 1 else ''
            accomm_availability = HTML(
                f'<p>The cost is {from_}{cheapest} '
                f'{settings.DEBCONF_BILLING_CURRENCY}/night for attendees who '
                f'do not receive a bursary. '
                f'Make sure to check "I want to apply for an accommodation '
                f'bursary" on the bursary page if you need one.</p>')
        else:
            accomm_availability = HTML(
                '<p>Conference-organised accommodation is only available to '
                'attendees who receive an accommodation bursary. If you are '
                'paying for your own accommodation, see our '
                '<a href="/about/accommodation">list of nearby options</a>.')

        option = ()
        if len(settings.PRICES['accomm']) > 1:
            self.fields['option'].choices = accomm_option_choices(
                bursary=self.bursary.request_accommodation,
                paid=not self.bursary.request_accommodation)
            option = (Field('option'),)
        if self.orga_nights:
            self.fields['nights'].choices = night_choices(orga=True)

        if PAID_ACCOMMODATION or self.bursary.request_accommodation:
            self.helper.layout = Layout(
                accomm_availability,
                Field('accomm', id='accomm'),
                Fieldset(
                    'Accommodation Details',
                    NightSelectionField('nights'),
                    *option,
                    Field('requirements'),
                    Field('family_usernames'),
                    css_id='accomm-details',
                )
            )
        else:
            self.helper.layout = Layout(accomm_availability)

    def clean(self):
        cleaned_data = super().clean()

        if not cleaned_data.get('accomm'):
            if self.bursary.request_accommodation:
                self.add_error('accomm',
                    'Accommodation bursary was requested, but no accommodation '
                    'selected.')
            return cleaned_data

        if not PAID_ACCOMMODATION and not self.bursary.request_accommodation:
            self.add_error(
                'accomm',
                'Accommodation is only available to attendees who receive an '
                'accommodation bursary.')

        if not cleaned_data.get('nights'):
            self.add_error(
                'accomm',
                'Please select the nights you require accommodation for.')

        return cleaned_data



from django.conf import settings
from django.db import migrations, models

from pytz import common_timezones


def tz_choices():
    """Duplicate logic from model, so that django doesn't constantly think
    choices has changed
    """
    timezones = common_timezones
    if settings.TIME_ZONE not in timezones:
        timezones.insert(0, settings.TIME_ZONE)
    return [(tz, tz) for tz in timezones]


class Migration(migrations.Migration):

    dependencies = [
        ('volunteers', '0003_fk_on_delete'),
    ]

    operations = [
        migrations.AddField(
            model_name='volunteer',
            name='timezone',
            field=models.TextField(default='Etc/UTC', max_length=32,
                                   choices=tz_choices()),
        ),
    ]



from django import forms
from django.conf import settings

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Field, Fieldset, HTML, Layout

from register.models import Attendee


INVOICE_GENERATED_ERROR = (
    'Your invoice has already been generated, '
    'contact the registration team if you need to update it.'
)
FINAL_DATES_ESTIMATE_LABEL = "Estimated, I haven't booked travel yet."
FINAL_DATES_FINAL_LABEL = 'Final, I have booked my travel.'


class ConferenceRegistrationForm(forms.Form):
    coc_ack = forms.BooleanField(
        label='I have read and promise to abide by the '
              '<a href="https://www.debconf.org/codeofconduct.shtml" '
              'target="_blank">'
              'DebConf Code of Conduct</a>',
        required=True,
    )
    fee = forms.ChoiceField(
        label='My registration fee',
        choices=Attendee.FEES.items(),
        help_text='We encourage attendees to pay for their attendance if they '
                  'can afford to do so.',
        widget=forms.RadioSelect,
        initial='pro',
        required=False,
    )
    arrival = forms.DateTimeField(
        label='I arrive at the venue on',
        help_text="Please estimate, if you haven't booked tickets, yet, "
                  'and update it when you have final dates.',
        required=False,
    )
    departure = forms.DateTimeField(
        label='I depart from the venue on',
        help_text="The date picker won't let you go beyond 10am on the "
                  "departure day.",
        required=False,
    )
    final_dates = forms.BooleanField(
        label='My dates are',
        widget=forms.Select(choices=(
                (False, FINAL_DATES_ESTIMATE_LABEL),
                (True, FINAL_DATES_FINAL_LABEL),
        )),
        initial=False,
        help_text="We'd like a rough indication of dates, even if you aren't "
                  'sure about the details yet. It helps us to plan.',
        required=False,
    )
    reconfirm = forms.BooleanField(
        label='I confirm my attendance',
        help_text="If you do not select this by July 16th, we'll assume you "
                  "aren't coming. If you are a paying attendee, checking "
                  "this box will generate your invoice.",
        required=False,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = False
        self.helper.include_media = False

        dates_list = []
        for label, start, end in settings.DEBCONF_DATES:
            dates_list.append('<dt>{}:</dt>'.format(label))
            if start == end:
                dates_list.append('<dd>{}</dd>'.format(start))
            else:
                dates_list.append('<dd>{} to {}</dd>'.format(start, end))

        if settings.DEBCONF_ONLINE:
            del self.fields['arrival']
            del self.fields['departure']
            del self.fields['final_dates']
            self.helper.layout = Layout(
                'coc_ack',
                'fee',
            )
        else:
            self.helper.layout = Layout(
                'coc_ack',
                'fee',
                Fieldset(
                    'Dates',
                    HTML('<dl>' + ''.join(dates_list) + '</dl>'
                         '<p>'
                         '  <a href="/schedule/important-dates/">'
                         '    Important Dates'
                         '  </a>'
                         '</p>'),
                    Field('arrival', id='arrival'),
                    Field('departure', id='departure'),
                    'final_dates',
                ),
            )
        if settings.RECONFIRMATION:
            self.helper.layout.append('reconfirm')

    def clean(self):
        cleaned_data = super().clean()

        if not settings.DEBCONF_ONLINE:
            if cleaned_data.get('final_dates'):
                for field in ('arrival', 'departure'):
                    if not cleaned_data.get(field):
                        self.add_error(
                            field, 'If your dates are final, please provide them')

            else:
                if cleaned_data.get('reconfirm'):
                    self.add_error(
                        'final_dates', 'Dates need to be final, to reconfirm')



from django.conf import settings
from django.contrib import messages

from bursary.models import Bursary
from register.forms.bursary import BursaryForm
from register.models.accommodation import Accomm, PAID_ACCOMMODATION
from register.views.core import RegisterStep


class BursaryView(RegisterStep):
    template_name = 'register/page/bursary.html'
    title = 'Bursary'
    form_class = BursaryForm

    def get_initial(self):
        user = self.request.user
        initial = {}

        try:
            bursary = user.bursary
        except Bursary.DoesNotExist:
            pass
        else:
            for field in bursary._meta.get_fields():
                if field.is_relation:
                    continue
                initial[field.name] = getattr(bursary, field.name)

        return initial

    def form_valid(self, form):
        user = self.request.user
        data = form.cleaned_data

        if not (data['request_food'] or data['request_accommodation']
                or data['request_travel'] or data['request_expenses']):
            # Update but don't create:
            Bursary.objects.filter(user=user).update(**data)
        elif settings.BURSARIES_CLOSED:
            # Only decrease existing requests:
            try:
                bursary = Bursary.objects.get(user=user)
            except Bursary.objects.DoesNotExist:
                pass
            else:
                bursary.travel_bursary = min(
                    bursary.travel_bursary or 0,
                    data['travel_bursary'] or 0) or None
                for field in ('request_travel', 'request_food',
                              'request_accommodation'):
                    if getattr(bursary, field) and not data[field]:
                        setattr(bursary, field, data[field])
                bursary.save()
        else:
            user.bursary, created = Bursary.objects.update_or_create(
                user=user, defaults=data)

        if not PAID_ACCOMMODATION:
            if not data['request_accommodation']:
                try:
                    user.attendee.accomm.delete()
                except Accomm.DoesNotExist:
                    pass
                else:
                    messages.warning(
                        self.request,
                        "Your accommodation registration is now cancelled. "
                        "You'll need to organise your own accommodation.")

        return super().form_valid(form)



from datetime import date

# A basic set of settings that is applicable for both test suites.
# Many of these will be overriden by conftest.py in the suites.

from debconf.common_settings import *

INSTALLED_APPS = (
    # minidebconf:
    'minidebconf',
    # debconf:
    'badges',
    'debconf',
    'exports',
    'bursary',
    'front_desk',
    'invoices',
    'register',
) + INSTALLED_APPS

# For MiniDebConf Tests:

# For DebConf Tests:
DEBCONF_BREAKFAST = True
DEBCONF_NAME = 'Test DebConf'
DEBCONF_DATES = (
    ('DebCamp', date(2003, 7, 12), date(2003, 7, 17)),
    ('DebConf', date(2003, 7, 18), date(2003, 7, 20)),
)
DEBCONF_LOCAL_CURRENCY_RATE = 1
DEBCONF_T_SHIRT_SIZES = (
    ('', 'No T-shirt'),
    ('s', 'Small'),
    ('l', 'Large'),
)
DEBCONF_SHOE_SIZES = (
    ('', 'No Shoes'),
    ('s', 'Small'),
    ('l', 'Large'),
)

INVOICE_PREFIX = 'DCTEST-'
PRICES = {
    'fee': {
        '': {
            'name': 'Regular',
            'price': 0,
        },
        'pro': {
            'name': 'Professional',
            'price': 10,
        },
    },
    'meal': {
        'breakfast': {
            'price': 1,
        },
        'lunch': {
            'price': 3,
        },
        'dinner': {
            'price': 5,
        },
        'conference_dinner': {
            'price': 7,
        },
    },
    'accomm': {},
}
STRIPE_SECRET_KEY = ''



from django.conf.urls import url

from badges.views import OwnBadge, CheckInBadgeView

urlpatterns = [
    url(r'^$', OwnBadge.as_view(),
        name='badges.own'),
    url(r'^check_in/(?P<username>[\w.@+-]+)/$', CheckInBadgeView.as_view(),
        name='badges.check_in'),
]



import requests

from django.core.management.base import BaseCommand

from wafer.talks.models import Talk


class Command(BaseCommand):
    help = 'Import published talk videos'

    def handle(self, *args, **options):
        released = requests.get('https://sreview.debian.net/released').json()
        for video in released:
            id_ = int(video['waferurl'].split('/')[2])
            url = video['publicurl']
            talk = Talk.objects.get(pk=id_)
            tu, created = talk.urls.update_or_create(
                description='Video', defaults={'url': url})

            if created:
                print('Set URL for {}'.format(talk.title))



from django.test import Client
from minidebconf.models import Registration
from minidebconf.forms import RegisterForm
from wafer.schedule.models import ScheduleBlock
from pytest import fixture


@fixture
def days(db):
    day1 = ScheduleBlock.objects.create(start_time='2020-06-01 09:00+00', end_time='2020-06-01 18:00+00')
    day2 = ScheduleBlock.objects.create(start_time='2020-06-02 09:00+00', end_time='2020-06-02 18:00+00')
    day3 = ScheduleBlock.objects.create(start_time='2020-06-03 09:00+00', end_time='2020-06-03 18:00+00')
    return [day1, day2, day3]



class TestRegistrationModel:

    def test_full_name(self, user):
        registration = Registration(user=user)
        assert registration.full_name == "Firstname Lastname"


class TestRegistrationForm:

    def test_basics(self, user, days):
        form = RegisterForm(
            instance=Registration(user=user),
            initial={"days": days}
        )
        assert not form.errors

    def test_get_full_name_from_user(self, user, days):
        registration = Registration.objects.create(user=user, country="BR")
        registration.days.add(days[0])
        form = RegisterForm(instance=registration)
        assert form.fields['full_name'].initial == registration.full_name

    def test_pass_full_name_on_to_user(self, user, days):
        form = RegisterForm(
            instance=Registration(user=user),
            data={"days": days, "country": "BR", "full_name": 'Foo Bar'},
        )
        assert not form.errors
        form.save()
        user.refresh_from_db()
        assert user.first_name == "Foo"
        assert user.last_name == "Bar"


class TestRegistrationView:
    def test_register(self, user, client, days):
        response = client.post(
            '/register/',
            {
                "days": [days[1].id, days[2].id],
                "country": "BR",
                "full_name": "Foo Bar",
            }
        )
        assert response.status_code == 302
        registration = Registration.objects.last()
        assert registration.user == user
        assert days[0] not in registration.days.all()
        assert days[1] in registration.days.all()
        assert days[2] in registration.days.all()

    def test_update_registration(self, user, client, days):
        record = Registration.objects.create(user=user, country="BR")
        record.days.add(days[0])
        response = client.post(
            '/register/',
            {
                "days": [days[1].id],
                "country": "FR",
                "full_name": "Baz Qux",
            }
        )
        assert response.status_code == 302
        registration = Registration.objects.get(user=user)
        assert registration.days.get() == days[1]
        assert registration.country == "FR"

    def test_unauthenticated(self, db):
        client = Client()
        response = client.get("/register/")
        assert response.status_code == 302
        assert response["Location"].startswith("/accounts/login/")


class TestCancelRegistrationView:
    def test_cancel_registration(self, user, client):
        record = Registration.objects.create(user=user)
        response = client.post('/unregister/')
        assert response.status_code == 302
        assert Registration.objects.filter(user=user).count() == 0

    def test_cancel_registration_get(self, user, client):
        record = Registration.objects.create(user=user)
        response = client.get('/unregister/')
        assert response.status_code == 200



from django.conf.urls import include, url
from minidebconf.views import RegisterView, UnregisterView, RegistrationFinishedView

urlpatterns = [
    url(r'^register/$', RegisterView.as_view(), name='register'),
    url(r'^unregister/$', UnregisterView.as_view(), name='unregister'),
    url(r'^register/finished/$', RegistrationFinishedView.as_view(), name='registration_finished'),
    url(r'', include('debconf.urls')),
    url(r'', include('wafer.urls')),
]
